@@ -1,22 +1,34 @@
-use crate::lexer::{Token, TokenIter};
-use crate::ast::{Ast, Expr};
-use crate::cmd;
-use crate::cmd::Command;
-use std::collections::{HashSet, HashMap};
-use std::io::{BufRead, BufReader};
-use std::fs::File;
+use crate::{
+    lexer::{Token, TokenIter},
+    ast::{Ast, Expr},
+    cmd,
+    cmd::Command,
+};
+use std::{
+    collections::HashMap,
+    io::{BufRead, BufReader},
+    fs::File,
+};
 
 /// Our hand-written parser.
 /// Use with parse().
 ///
 pub struct Parser {
     symbol_table: HashMap<String, Ast>,
+    pause: bool,
+    step: bool,
+    count_steps: bool,
+    non_interactive_mode: bool,
 }
 
 impl Parser {
     pub fn new() -> Parser {
         Parser {
             symbol_table: HashMap::new(),
+            pause: false,
+            step: true,
+            count_steps: false,
+            non_interactive_mode: false,
         }
     }
 
@@ -31,7 +43,7 @@ impl Parser {
     /// ```
     /// # use lambda_calc::Parser;
     /// let mut parser = Parser::new();
-    /// assert_eq!(parser.parse("a b c d"), parser.parse("(((a b) c) d)"));
+    /// assert_eq!(parser.parse("a b c d", None), parser.parse("(((a b) c) d)", None));
     /// ```
     ///
     /// - The lambda body stretches as far as possible:
@@ -39,14 +51,14 @@ impl Parser {
     /// ```
     /// # use lambda_calc::Parser;
     /// let mut parser = Parser::new();
-    /// assert_eq!(parser.parse("lambda x . a b c"), parser.parse("(lambda x . a b c)"));
+    /// assert_eq!(parser.parse("lambda x . a b c", None), parser.parse("(lambda x . a b c)", None));
     /// ```
     ///
     /// ```
     /// # use lambda_calc::Parser;
     /// let mut parser = Parser::new();
-    /// assert_eq!(parser.parse("lambda x . a lambda y . y"),
-    ///     parser.parse("(lambda x . (a (lambda y . y)))"));
+    /// assert_eq!(parser.parse("lambda x . a lambda y . y", None),
+    ///     parser.parse("(lambda x . (a (lambda y . y)))", None));
     /// ```
     ///
     /// - Unnecessary parentheses are ignored:
@@ -54,7 +66,7 @@ impl Parser {
     /// ```
     /// # use lambda_calc::Parser;
     /// let mut parser = Parser::new();
-    /// assert_eq!(parser.parse("((((a))))"), parser.parse("a"));
+    /// assert_eq!(parser.parse("((((a))))", None), parser.parse("a", None));
     /// ```
     ///
     /// Additionally,
@@ -64,23 +76,57 @@ impl Parser {
     /// ```
     /// # use lambda_calc::Parser;
     /// let mut parser = Parser::new();
-    /// assert_eq!(parser.parse("lambda x y z . x y"),
-    ///     parser.parse("lambda x . lambda y . lambda z . x y"));
+    /// assert_eq!(parser.parse("lambda x y z . x y", None),
+    ///     parser.parse("lambda x . lambda y . lambda z . x y", None));
     /// ```
     ///
     ///
-    /// - Lambda terms may be bound to symbols and later used:
+    /// - Lambda terms may be bound to symbols and later used.
+    ///
+    /// Note that the symbol substitution is lazy, meaning it is deferred until
+    /// truly needed when perfoming a beta reduction:
+    ///
+    /// ```
+    /// # use lambda_calc::{Parser, Ast, ast::Expr};
+    /// let mut parser = Parser::new();
+    /// assert_eq!(parser.parse("K = lambda x y . x", None), None);
+    /// let redex = parser.parse("K a b", None);
+    /// // we haven't asked for any beta reduction, so K has not been
+    /// // substituted yet
+    /// let expected = Ast::new(Expr::Redex(
+    ///     Box::new(Ast::new(Expr::Redex(
+    ///         Box::new(Ast::new(Expr::Var { name: "K".to_string(), is_free: true })),
+    ///         Box::new(Ast::new(Expr::Var { name: "a".to_string(), is_free: true })),
+    ///     ))),
+    ///     Box::new(Ast::new(Expr::Var { name: "b".to_string(), is_free: true })),
+    /// )); // ((K a) b)
+    /// assert_eq!(redex, Some(expected));
+    /// let reduced = redex.unwrap().beta_reduce_quiet(&parser);
+    /// let expected = Ast::new(Expr::Var { name: "a".to_string(), is_free: true });
+    /// assert_eq!(reduced, expected);
+    /// ```
+    ///
+    /// One must take care when binding an expression with free variables;
+    /// because of the lazy substitution, these variables might misbehave when
+    /// captured, e.g. in
     ///
     /// ```
     /// # use lambda_calc::Parser;
     /// let mut parser = Parser::new();
-    /// assert_eq!(parser.parse("K = lambda x y . x"), None);
-    /// assert_eq!(parser.parse("I = lambda x . x"), None);
-    /// assert_eq!(parser.parse("K I"),
-    ///     parser.parse("(lambda x y . x) (lambda x . x)"));
+    /// assert_eq!(parser.parse("foo = \\f -> f x", None), None);
+    /// let reduced = parser.parse("(\\x -> foo) a", None)
+    ///     .unwrap()
+    ///     .beta_reduce_quiet(&parser);
+    /// // the "x" in foo's definition is never replaced with "a",
+    /// // because the definition is only substituted later
+    /// let expected = parser.parse("foo", None)
+    ///     .unwrap()
+    ///     .beta_reduce_quiet(&parser);
+    /// assert_eq!(reduced, expected);
     /// ```
     ///
-    pub fn parse(&mut self, line: &str) -> Option<Ast> {
+
+    pub fn parse(&mut self, line: &str, file_info: Option<String>) -> Option<Ast> {
         let token_iter = TokenIter::new(&line);
         {
             let mut new_token_iter = token_iter.clone();
@@ -89,18 +135,55 @@ impl Parser {
                 return None;
             }
         }
-        LineParser::new(token_iter).parse(&mut self.symbol_table)
+        LineParser::new(token_iter, file_info).parse(self)
     }
 
     /// Parse all lines in filename.
     pub fn parse_file(&mut self, filename: &str) -> std::io::Result<()> {
         let reader = BufReader::new(File::open(filename)?);
-        for line in reader.lines() {
-            self.parse(&line?);
+
+        for (line_num, line) in reader.lines().enumerate() {
+            let file_info = format!("file '{}', line {}: ", filename, line_num + 1);
+            if let Some(ast) = self.parse(&line?, Some(file_info)) {
+                if self.non_interactive_mode {
+                    println!("{:#}", ast.beta_reduce_quiet(&self));
+                }
+            }
         }
         Ok(())
     }
 
+    /// Return an immutable reference to an expression if its name can be found
+    /// in the symbol table.
+    pub fn get_symbol(&self, name: &str) -> Option<&Ast> {
+        self.symbol_table.get(name)
+    }
+
+    /// Insert a symbol and its corresponding expression into the symbol table.
+    pub fn insert_symbol(&mut self, name: &str, ast: Ast) {
+        self.symbol_table.insert(name.to_string(), ast);
+    }
+
+    /// Return whether pause mode is on or off.
+    pub fn pause(&self) -> bool {
+        self.pause
+    }
+
+    /// Return whether step mode is on or off.
+    pub fn step(&self) -> bool {
+        self.step
+    }
+
+    /// Return whether the count step mode is on or off.
+    pub fn count_steps(&self) -> bool {
+        self.count_steps
+    }
+
+    /// Turn on non-interactive mode.
+    pub fn set_non_interactive_mode(&mut self) {
+        self.non_interactive_mode = true;
+    }
+
     fn run_command(&mut self, cmd: Command, arg: Option<Token>) {
         match cmd {
             Command::Help => {
@@ -119,6 +202,9 @@ impl Parser {
                 _ => panic!("lexer should have returned an Id as the argument to ':load'"),
             },
             Command::Define => panic!("Command::Define should never be returned by the lexer!"),
+            Command::Pause => handle_bool_cmd(&mut self.pause, arg),
+            Command::Step => handle_bool_cmd(&mut self.step, arg),
+            Command::CountSteps => handle_bool_cmd(&mut self.count_steps, arg),
         }
     }
 }
@@ -128,20 +214,22 @@ struct LineParser<'a, I>
     where I: Iterator<Item = Token<'a>> + Clone
 {
     token_iter: I,
-    lambda_vars: HashSet<String>,
+    file_info: Option<String>,
+    lambda_vars: Vec<String>,
 }
 
 impl<'a, I> LineParser<'a, I>
     where I: Iterator<Item = Token<'a>> + Clone
 {
-    fn new(token_iter: I) -> LineParser<'a, I> {
+    fn new(token_iter: I, file_info: Option<String>) -> LineParser<'a, I> {
         LineParser {
             token_iter,
-            lambda_vars: HashSet::new(),
+            file_info,
+            lambda_vars: Vec::new(),
         }
     }
 
-    fn parse(&mut self, symbol_table: &mut HashMap<String, Ast>) -> Option<Ast> {
+    fn parse(&mut self, parser: &mut Parser) -> Option<Ast> {
         if let None = self.token_iter.clone().next() {
             return None; // empty line, ignore.
         }
@@ -150,19 +238,15 @@ impl<'a, I> LineParser<'a, I>
             return None;
         }
         if is_def {
-            self.parse_def(symbol_table);
+            self.parse_def(parser);
             None
         } else {
-            let mut ast = self.parse_ast(Vec::new())?;
-            {
-                let mut lambda_vars = HashSet::new();
-                ast.substitute_symbols_from(symbol_table, &mut lambda_vars);
-            }
+            let ast = self.parse_ast(Vec::new())?;
             Some(ast)
         }
     }
 
-    fn parse_def(&mut self, symbol_table: &mut HashMap<String, Ast>) {
+    fn parse_def(&mut self, parser: &mut Parser) {
         let name = match self.token_iter.next() {
             Some(Token::Id(s)) => s,
             _ => panic!("expected definition, but 1st token is not an ID"),
@@ -172,25 +256,24 @@ impl<'a, I> LineParser<'a, I>
             _ => panic!("expected definition, but 2nd token is not '=' or ':='"),
         }
         match self.parse_ast(Vec::new()) {
-            None => eprintln!("a definition can't bind to an empty expression"),
-            Some(mut ast) => match ast.expr_ref() {
+            None => self.print_err_msg("a definition can't bind to an empty expression"),
+            Some(ast) => match ast.expr_ref() {
                 Expr::LambdaTerm { var_name: _, body: _ } => {
-                    {
-                        let mut lambda_vars = HashSet::new();
-                        ast.substitute_symbols_from(symbol_table, &mut lambda_vars);
-                    }
-                    symbol_table.insert(name.to_string(), ast);
+                    parser.insert_symbol(name,  ast);
                 },
                 Expr::Redex(_,_) => {
-                    {
-                        let mut lambda_vars = HashSet::new();
-                        ast.substitute_symbols_from(symbol_table, &mut lambda_vars);
-                    }
-                    let non_redex = ast.beta_reduce_quiet();
-                    symbol_table.insert(name.to_string(), non_redex);
+                    let non_redex = ast.beta_reduce_quiet(parser);
+                    parser.insert_symbol(name, non_redex);
                 },
-                Expr::Var{name: _, is_free: _} => {
-                    eprintln!("a definition can't bind to a single variable");
+                Expr::Var{name: var_name, is_free, } => {
+                    assert!(is_free);
+                    if let Some(ast) = parser.get_symbol(&var_name) {
+                        let ast = ast.clone();
+                        // bind to the symbol's definition directly
+                        parser.insert_symbol(name, ast);
+                    } else {
+                        self.print_err_msg("a definition can't bind to a single variable (unless it also has a definition)");
+                    }
                 }
             }
         }
@@ -211,12 +294,15 @@ impl<'a, I> LineParser<'a, I>
                     );
                 },
                 Some(Token::Id(s)) => {
-                    let is_free = !self.lambda_vars.contains(s);
+                    let is_free = !self.lambda_vars.iter().any(|x| x == s);
                     queue.push(Ast::new(Expr::Var { name: s.to_string(), is_free, }));
                 },
                 Some(Token::Lambda) => {
-                    queue.push(self.parse_lambda());
-                    self.lambda_vars.clear();
+                    let (ast, var_count) = self.parse_lambda();
+                    queue.push(ast);
+                    for _ in 0..var_count {
+                        self.lambda_vars.pop();
+                    }
                     return finalize_redex(queue);
                 },
                 t @ _ => {
@@ -235,26 +321,34 @@ impl<'a, I> LineParser<'a, I>
     // which is treated as equivalent to
     //  lambda x -> (lambda y -> x)
     //
-    fn parse_lambda(&mut self) -> Ast {
+    // The number returned is the number of lambda variables pushed into
+    // self.lambda_vars.
+    //
+    fn parse_lambda(&mut self) -> (Ast, i32) {
         match self.token_iter.next() {
             Some(Token::Id(name)) => {
-                if self.lambda_vars.contains(name) {
+                if self.lambda_vars.iter().any(|x| x == name) {
                     // Ideally we should do something less drastic than panicking,
                     // e.g. returning None, but that could lead to a non-None
                     // inconsistent expression, which is even worse.
                     //
                     panic!("outer lambda variable cannot appear again as an inner lambda variable");
                 }
-                self.lambda_vars.insert(name.to_string());
-                Ast::new(Expr::LambdaTerm {
+                self.lambda_vars.push(name.to_string());
+
+                let (lambda_body, var_count) = self.parse_lambda();
+                let ast = Ast::new(Expr::LambdaTerm {
                     var_name: name.to_string(),
-                    body: Box::new(self.parse_lambda()),
-                })
+                    body: Box::new(lambda_body),
+                });
+
+                (ast, var_count + 1)
             },
             Some(Token::Gives) => {
                 // finalize lambda term by returning its body.
-                self.parse_ast(Vec::new())
-                    .expect("lambda term can't have empty body")
+                let ast = self.parse_ast(Vec::new())
+                    .expect("lambda term can't have empty body");
+                (ast, 0)
             },
             _ => panic!("this token is not allowed in the head of a lambda term."),
         }
@@ -277,16 +371,15 @@ impl<'a, I> LineParser<'a, I>
         for (i, token) in token_iter.enumerate() {
             match token {
                 Token::Invalid => {
-                    eprintln!("token number {} is invalid", i);
+                    self.print_err_msg(&format!("token number {} is invalid", i));
                     return false;
                 },
                 Token::OpenParen => paren_count += 1,
                 Token::CloseParen => paren_count -= 1,
                 Token::Def => {
                     if !is_def {
-                        eprintln!(
-                            "wrong syntax for definition, should be <var> = <expr> (token {})",
-                            i
+                        self.print_err_msg(
+                            &format!("wrong syntax for definition, should be <var> = <expr> (token {})", i)
                         );
                         return false;
                     }
@@ -295,14 +388,21 @@ impl<'a, I> LineParser<'a, I>
             }
         }
         if paren_count > 0 {
-            eprintln!("{} unclosed parentheses", paren_count);
+            self.print_err_msg(&format!("{} unclosed parentheses", paren_count));
             return false;
         } else if paren_count < 0 {
-            eprintln!("{} extra closing parentheses", -paren_count);
+            self.print_err_msg(&format!("{} extra closing parentheses", -paren_count));
             return false;
         }
         true
     }
+
+    fn print_err_msg(&self, s: &str) {
+        if let Some(fs) = &self.file_info {
+            eprint!("{}", fs);
+        }
+        eprintln!("{}", s);
+    }
 }
 
 fn finalize_redex(mut queue: Vec<Ast>) -> Option<Ast> {
@@ -316,6 +416,24 @@ fn finalize_redex(mut queue: Vec<Ast>) -> Option<Ast> {
     Some(result)
 }
 
+fn handle_bool_cmd(val: &mut bool, arg: Option<Token>) {
+    match arg {
+        None => eprintln!("command requires an argument on|off|toggle"),
+        Some(Token::Id(s)) => {
+            if s == "on" {
+                *val = true;
+            } else if s == "off" {
+                *val = false;
+            } else if s == "toggle" {
+                *val = !*val;
+            } else {
+                eprintln!("expected 'on', 'off' or 'toggle', found '{}'", s);
+            }
+        },
+        _ => eprintln!("invalid argument: must be 'on', 'off' or 'toggle'"),
+    }
+}
+
 #[cfg(test)]
 mod tests {
     use super::*;
@@ -360,15 +478,15 @@ mod tests {
 
     #[test]
     fn empty_expr() {
-        let mut symbol_table = HashMap::<String, Ast>::new();
-        let expr = LineParser::new(vec![].into_iter()).parse(&mut symbol_table);
+        let mut parser = Parser::new();
+        let expr = LineParser::new(vec![].into_iter(), None).parse(&mut parser);
         assert_eq!(None, expr);
     }
 
     // skeleton for non-empty expressions' tests (not definitions)
     fn expr_test<'a>(tokens: Vec<Token<'a>>, expected: Ast) {
-        let mut symbol_table = HashMap::<String, Ast>::new();
-        let ast = LineParser::new(tokens.into_iter()).parse(&mut symbol_table);
+        let mut parser = Parser::new();
+        let ast = LineParser::new(tokens.into_iter(), None).parse(&mut parser);
         assert_eq!(Some(expected), ast);
     }
 
@@ -395,18 +513,6 @@ mod tests {
         ]; // x y z w       -- should become (((x y) z) w)
         expr_test(
             tokens,
-            /*
-            Expr::Redex(
-                Box::new(Expr::Redex(
-                    Box::new(Expr::Redex(
-                        Box::new(Expr::Var{ name: "x".to_string(), is_free: true }),
-                        Box::new(Expr::Var{ name: "y".to_string(), is_free: true }),
-                    )),
-                    Box::new(Expr::Var{ name: "z".to_string(), is_free: true }),
-                )),
-                Box::new(Expr::Var{ name: "w".to_string(), is_free: true }),
-            )
-            */
             redex_no_box(
                 redex(
                     redex(free_var("x"), free_var("y")),
@@ -433,18 +539,6 @@ mod tests {
 
         expr_test(
             tokens,
-            /*
-            Expr::Redex(
-                Box::new(Expr::Redex(
-                    Box::new(Expr::Redex(
-                        Box::new(Expr::Var{ name: "x".to_string(), is_free: true }),
-                        Box::new(Expr::Var{ name: "y".to_string(), is_free: true }),
-                    )),
-                    Box::new(Expr::Var{ name: "z".to_string(), is_free: true }),
-                )),
-                Box::new(Expr::Var{ name: "w".to_string(), is_free: true }),
-            )
-            */
             redex_no_box(
                 redex(
                     redex(free_var("x"), free_var("y")),
@@ -530,15 +624,6 @@ mod tests {
         ]; // lambda x . (x a)
         expr_test(
             tokens,
-            /*
-            Expr::LambdaTerm {
-                var_name: "x".to_string(),
-                body: Box::new(Expr::Redex(
-                        Box::new(Expr::Var{ name: "x".to_string(), is_free: false }),
-                        Box::new(Expr::Var{ name: "a".to_string(), is_free: true }),
-                ))
-            }
-            */
             lambda_no_box("x", redex(bound_var("x"), free_var("a")))
         );
     }
@@ -568,25 +653,6 @@ mod tests {
                                   bound_var("z")
                               ),
             )))
-            /*
-            Expr::LambdaTerm {
-                var_name: "x".to_string(),
-                body: Box::new(Expr::LambdaTerm {
-                    var_name: "y".to_string(),
-                    body: Box::new(Expr::LambdaTerm {
-                        var_name: "z".to_string(),
-                        body:
-                            Box::new(Expr::Redex(
-                                Box::new(Expr::Redex(
-                                    Box::new(Expr::Var{ name: "x".to_string(), is_free: false }),
-                                    Box::new(Expr::Var{ name: "y".to_string(), is_free: false }),
-                                )),
-                            Box::new(Expr::Var{ name: "z".to_string(), is_free: false }),
-                        ))
-                    })
-                })
-            }
-            */
         );
     }
 
@@ -938,15 +1004,46 @@ mod tests {
         );
     }
 
+    #[test]
+    fn bound_vars_inside_multiple_lambdas() {
+        let tokens = vec![
+            Token::Lambda, Token::Id("f"), Token::Gives,
+            Token::OpenParen,
+            Token::Lambda, Token::Id("x"), Token::Gives, Token::Id("f"),
+            Token::CloseParen,
+            Token::OpenParen,
+            Token::Lambda, Token::Id("x"), Token::Gives, Token::Id("f"),
+            Token::CloseParen,
+        ]; // \f -> (\x -> f) (\x -> f)
+        expr_test(
+            tokens,
+            lambda_no_box(
+                "f",
+                redex(
+                    lambda("x", bound_var("f")),
+                    lambda("x", bound_var("f")),
+                ),
+            ),
+        );
+    }
+
+    #[test]
+    fn insert_get_symbol() {
+        let mut parser = Parser::new();
+        let ast = lambda_no_box("x", bound_var("x"));
+        parser.insert_symbol("I", ast.clone());
+        assert_eq!(parser.get_symbol("I"), Some(&ast));
+    }
+
     // for definition tests.
-    fn def_test<'a>(def_tokens: Vec<Token<'a>>, tokens: Vec<Token<'a>>, expected: Ast) {
-        let mut symbol_table = HashMap::<String, Ast>::new();
+    fn def_test<'a>(name: &str, def_tokens: Vec<Token<'a>>, expected: Option<&Ast>) {
+        let mut parser = Parser::new();
         assert_eq!(
             None,
-            LineParser::new(def_tokens.into_iter()).parse(&mut symbol_table)
+            LineParser::new(def_tokens.into_iter(), None).parse(&mut parser)
         );
-        let ast = LineParser::new(tokens.into_iter()).parse(&mut symbol_table);
-        assert_eq!(Some(expected), ast);
+        let ast = parser.get_symbol(name);
+        assert_eq!(expected, ast);
     }
 
     #[test]
@@ -955,9 +1052,8 @@ mod tests {
             Token::Id("I"), Token::Def,
             Token::Lambda, Token::Id("x"), Token::Gives, Token::Id("x")
         ];
-        let tokens = vec![ Token::Id("I") ];
         let expected = lambda_no_box("x", bound_var("x"));
-        def_test(def_tokens, tokens, expected);
+        def_test("I", def_tokens, Some(&expected));
     }
 
     #[test]
@@ -965,123 +1061,30 @@ mod tests {
         let def_tokens = vec![
             Token::Id("x"), Token::Def, Token::Id("a")
         ];
-        let tokens = vec![ Token::Id("x") ];
-        let expected = free_var_no_box("x");
-        def_test(def_tokens, tokens, expected);
+        def_test("x", def_tokens, None);
     }
 
     #[test]
-    fn def_replaced_in_free_lambda_var() {
-        let def_tokens = vec![
-            Token::Id("I"), Token::Def,
-            Token::Lambda, Token::Id("x"), Token::Gives, Token::Id("x")
-        ];
-        let tokens = vec![
-            Token::Lambda, Token::Id("y"), Token::Gives,
-            Token::Id("I"), Token::Id("y"),
-        ];
-        let expected = lambda_no_box(
-            "y",
-            redex(
-                lambda("x",
-                       bound_var("x"),
-                ),
-                bound_var("y"),
-            ),
-        );
-        def_test(def_tokens, tokens, expected);
-    }
-
-    #[test]
-    fn def_not_replaced_in_bound_var() {
-        let def_tokens = vec![
-            Token::Id("a"), Token::Def,
-            Token::Lambda, Token::Id("x"), Token::Gives, Token::Id("x")
-        ];
-        let tokens = vec![
-            Token::Lambda, Token::Id("a"), Token::Gives, Token::Id("a")
-        ];
-        let expected = lambda_no_box("a", bound_var("a"));
-        def_test(def_tokens, tokens, expected);
-    }
-
-    #[test]
-    fn def_replaced_redex() {
-        let def_tokens = vec![
-            Token::Id("I"), Token::Def,
-            Token::Lambda, Token::Id("x"), Token::Gives, Token::Id("x")
-        ];
-
-        let tokens = vec![
-            Token::Id("I"), Token::Id("I")
-        ];
-        let expected = redex_no_box(
-            lambda("x", bound_var("x")),
-            lambda("x", bound_var("x")),
-        );
-        def_test(def_tokens, tokens, expected);
-    }
-
-    #[test]
-    fn def_with_redex() {
-        let def_tokens = vec![
-            Token::Id("a"), Token::Def,
-            Token::OpenParen,
-            Token::Lambda, Token::Id("x"), Token::Gives, Token::Id("x"),
-            Token::CloseParen,
-            Token::Lambda, Token::Id("y"), Token::Gives, Token::Id("y"),
-        ];
-        let tokens = vec![ Token::Id("a") ];
-        let expected = lambda_no_box("y", bound_var("y"));
-        def_test(def_tokens, tokens, expected);
-    }
-
-    #[test]
-    fn multilevel_def() {
-        let def1_tokens = vec![
-            Token::Id("a"), Token::Def,
-            Token::Lambda, Token::Id("x"), Token::Gives, Token::Id("x"),
-        ]; // a = lambda x . x
-        let def2_tokens = vec![
-            Token::Id("b"), Token::Def,
-            Token::Lambda, Token::Id("y"), Token::Gives, Token::Id("a"),
-        ]; // b = lambda y . a
-        let tokens = vec![ Token::Id("b") ];
-
-        let mut symbol_table = HashMap::<String, Ast>::new();
-        assert_eq!(
-            LineParser::new(def1_tokens.into_iter()).parse(&mut symbol_table),
-            None
-        );
-        assert_eq!(
-            LineParser::new(def2_tokens.into_iter()).parse(&mut symbol_table),
-            None
-        );
-        let expr = LineParser::new(tokens.into_iter()).parse(&mut symbol_table);
-        assert_eq!(
-            Some(lambda_no_box("y", lambda("x", bound_var("x")))),
-            expr
-        );
-    }
-
-    #[test]
-    fn alpha_conversion_def() {
-        let def_tokens = vec![
-            Token::Id("id"), Token::Def,
+    fn symbol_alias() {
+        let def1 = vec![
+            Token::Id("f"), Token::Def,
             Token::Lambda, Token::Id("x"), Token::Gives, Token::Id("x"),
         ];
-        let tokens = vec![
-            Token::Lambda, Token::Id("x"), Token::Gives,
-            Token::Id("x"), Token::Id("id"),
+        let def2 = vec![
+            Token::Id("g"), Token::Def, Token::Id("f"),
         ];
-        let expected = lambda_no_box(
-            "x",
-            redex(
-                bound_var("x"),
-                lambda("x'", bound_var("x'")),
-            ),
+        let mut parser = Parser::new();
+        assert_eq!(
+            None,
+            LineParser::new(def1.into_iter(), None).parse(&mut parser)
+        );
+        assert_eq!(
+            None,
+            LineParser::new(def2.into_iter(), None).parse(&mut parser)
+        );
+        assert_eq!(
+            parser.get_symbol("g"),
+            Some(&lambda_no_box("x", bound_var("x"))),
         );
-        def_test(def_tokens, tokens, expected);
-
     }
 }
